# 2021 Leo Kaindl

"""
This module builds a multiple sequence alignment for each gene and concats
them into one :code:`FASTA` file as input for the :class:`raxml` module.
"""

import logging
import os
import random
import shutil
import subprocess
import sys
from os import path
from pathlib import Path
from time import time, sleep

from Bio import SeqIO

from ab12phylo_cmd.filter import chmod_x


class msa_build:
    """Builds a Multiple Sequence Alignment"""

    def __init__(self, args, seq_counts, no_run=False):
        self.log = logging.getLogger(__name__)
        self.dir = args.dir
        self.genes = args.genes
        self.algo = args.msa_algo
        self.email = args.user
        self.msa = args.msa
        self.sep = args.sep
        self.missing_samples = args.missing_samples
        self.tools_path = path.join(path.abspath(path.dirname(__file__)), 'tools')

        # look for pre-installed version of selected algorithm
        self.binary = shutil.which(args.msa_algo)

        if no_run:
            return

        # build MSAs
        for gene in self.genes:
            if self.binary is not None:
                self.build_local(gene)
            else:
                self.build_remote(gene)
                sleep(1)

            # trim MSAs using Gblocks
            self.trim_msa(gene, seq_counts[gene], args.gblocks)

        # concat MSAs
        self.concat_msa()

    def reset_paths(self, _dir, msa):
        self.dir = _dir
        self.msa = msa

    def build_local(self, gene, no_run=False, new_arg=False):
        """Build MSAs locally using a pre-installed binary"""
        log_file = path.join(self.dir, gene, self.algo + '.log')
        fasta = path.join(self.dir, gene, gene + '.fasta')
        raw_msa = path.join(self.dir, gene, gene + '_raw_msa.fasta')
        self.log.debug('preparing %s MSA run' % self.algo)

        if new_arg:
            arg = new_arg

        elif self.algo == 'mafft':
            arg = '%s --thread %d --auto  "%s" > "%s"' \
                  % (self.binary, os.cpu_count(), fasta, raw_msa)

        elif self.algo == 'clustalo':
            arg = '%s --in "%s" --out "%s" --outfmt fasta --threads %d --force --verbose --auto' \
                  % (self.binary, fasta, raw_msa, os.cpu_count())
            if sys.platform in ['win32', 'cygwin']:
                arg += ' & exit /b 0'

        elif self.algo == 'muscle':
            arg = '%s -in "%s" -out "%s"' \
                  % (self.binary, fasta, raw_msa)

        elif self.algo == 't_coffee':
            arg = '%s -in "%s" -out "%s" -output fasta_aln -type dna ' \
                  % (self.binary, fasta, raw_msa)
        else:
            assert False

        if no_run:
            return arg
        self._run(arg, log_file, 'pre-installed %s' % self.algo, no_exit=True)

    def build_remote(self, gene, no_run=False, new_arg=False):
        """Builds an MSA online using an EBI API client"""
        log_file = path.join(self.dir, gene, self.algo + '.log')
        fasta = path.join(self.dir, gene, gene + '.fasta')
        raw_msa = path.join(self.dir, gene, gene + '_raw_msa.fasta')

        self.log.warning('running %s online' % self.algo.upper())

        if new_arg:
            arg = new_arg
        else:
            if self.algo == 't_coffee':
                self.algo = 'tcoffee'

            # create base call
            arg = '%s "%s" --email %s --outfile "%s" --sequence "%s" ' \
                  % (sys.executable, path.join(self.tools_path, 'MSA_clients', self.algo + '.py'),
                     self.email, path.join(self.dir, gene, 'msa'), fasta)

            # adapt for specific algorithm
            if self.algo == 'mafft':
                arg += '--stype dna'
            elif self.algo == 'clustalo':
                arg += '--stype dna --outfmt fa'
            elif self.algo == 'muscle':
                arg += '--format fasta'
            elif self.algo == 'tcoffee':
                arg += '--stype dna --format fasta_aln'

        if no_run:
            return arg

        # build an MSA for each gene
        self._run(arg, log_file, 'online %s' % self.algo)
        try:
            shutil.move(path.join(self.dir, gene, 'msa.aln-fasta.fasta'), raw_msa)
        except FileNotFoundError:
            shutil.move(path.join(self.dir, gene, '%s_raw_msa.fasta' % gene), raw_msa)

    def trim_msa(self, gene, seq_count, gblocks_mode):
        """
        Trims an MSA using Gblocks, using a pre-installed or deployed version.

        :param gene: this helps find the right files to trim
        :param seq_count: for computing settings
        :param gblocks_mode: can be ['skip', 'relaxed', 'balanced', 'semi_strict', 'strict']
        :return:
        """
        log_file = path.join(self.dir, gene, 'gblocks.log')
        raw_msa = path.join(self.dir, gene, gene + '_raw_msa.fasta')

        if gblocks_mode == 'skip':
            shutil.copy(raw_msa, path.join(self.dir, gene, gene + '_msa.fasta'))
            self.log.info('skipped Gblocks trimming, only copied file')

        else:
            # look for local Gblocks
            binary = shutil.which('Gblocks')
            local = True
            if binary is None:
                # pick deployed Gblocks
                binary = path.join(self.tools_path, 'Gblocks_0.91b', 'Gblocks').replace(' ', '\ ')
                local = False
                if sys.platform in ['win32', 'cygwin']:
                    binary += '.exe'
            chmod_x(binary)

            # set Gblocks options
            b4 = 5
            if gblocks_mode == 'relaxed':
                # set the minimal permissible minimum number of identical
                # sequences per position to define a conserved position
                cons = seq_count // 2 + 1
                # minimal number for a conserved flanking position
                flank = cons
                # keep no, half or all gap positions
                gaps = ['n', 'h', 'a'][1]
                self.log.info('running relaxed Gblocks')

            elif gblocks_mode == 'balanced':
                cons = seq_count // 2 + 1
                flank = min(seq_count // 4 * 3 + 1, seq_count)
                gaps = ['n', 'h', 'a'][1]
                self.log.info('running balanced Gblocks')

            elif gblocks_mode == 'default':
                cons = seq_count // 2 + 1
                flank = min(int(seq_count * 0.85) + 1, seq_count)
                gaps = ['n', 'h', 'a'][0]
                b4 = 10
                self.log.info('running Gblocks at default settings')

            else:
                cons = int(seq_count * 0.9)
                flank = cons
                gaps = ['n', 'h', 'a'][0]
                self.log.info('running strict Gblocks')

            # create base call
            arg = '%s "%s" -t=d -b2=%d -b1=%d -b4=%d -b5=%s -e=".txt" -d=n -s=y -p=n' \
                  % (binary, Path(raw_msa).resolve(), flank, cons, b4, gaps)  # don't swap order!
            # force return code
            if sys.platform in ['win32', 'cygwin']:
                arg += ' & exit /b 0'
            else:
                arg += '; exit 0'
            # MARK the -t=d sets the mode to nucleotides ... adapt?
            self._run(arg, log_file, 'pre-installed Gblocks' if local else 'out-of-the-box Gblocks')
            shutil.move(raw_msa + '.txt', path.join(self.dir, gene, gene + '_msa.fasta'))

    def concat_msa(self, gui=False):
        """Reads all trimmed MSAs to memory, then iterates over samples, writes concatenated MSA"""
        self.log.debug('concatenating per-gene MSAs')
        # read in all MSAs using SeqIO
        all_records = {gene: {record.id: record.upper() for record in SeqIO.parse(
            path.join(self.dir, gene, gene + '_msa.fasta'), 'fasta')} for gene in self.genes}

        # get the length of the trimmed concat MSA
        msa_len = 0
        for gene in all_records.keys():
            msa_len += len(random.choice(list(all_records[gene].values())))

        missing_genes = {gene: list() for gene in self.genes}

        with open(self.msa, 'w') as msa:
            shared = 0
            # iterate over samples available for first gene
            _ids = gui if gui else set(all_records[self.genes[0]])
            for sample_id in _ids:
                # get SeqRecord for first gene
                record = all_records[self.genes[0]].pop(sample_id)

                skip = False
                # append other genes
                for gene in self.genes[1:]:
                    try:
                        record += self.sep  # to visually separate genes in the MSA
                        record += all_records[gene].pop(sample_id)
                    except KeyError:
                        missing_genes[gene].append(sample_id)
                        skip = True
                # write to file
                if not skip:
                    shared += 1
                    SeqIO.write(record, msa, 'fasta')
        if len(self.genes) > 1:
            if shared == 0:
                msg = 'No samples shared across all genes.'
                self.log.error(msg)
                raise ValueError(msg) if gui else exit(1)
            else:
                self.log.info('finished writing concat MSA with %d entries' % shared)
        else:
            if msa_len == 0:
                msg = 'No conserved sites found. Please try a more relaxed trimming mode!'
                self.log.error(msg)
                if not gui:
                    raise ValueError(msg) if gui else exit(1)
            self.log.info('copied MSA to result root')
        self.log.info('MSA shape: %dx%d' % (msa_len, shared))

        if gui:
            return msa_len, shared

        # any remaining samples were missing from first gene
        [missing_genes[self.genes[0]].extend(all_records[gene].keys()) for gene in self.genes[1:]]

        # write info about missing samples to file and log
        missing = {gene: ', '.join(set(samples)) for gene, samples
                   in missing_genes.items() if set(samples)}
        if missing:
            with open(self.missing_samples, 'w') as fh:
                fh.write('gene\tmissing samples\n')
                for k, v in missing.items():
                    fh.write('%s\t%s\n' % (k, v))
                    self.log.info('samples missing from %s: %s' % (k, v))

    def _run(self, arg, log_file, info, no_exit=False):
        """Runs a command and writes output to log-file"""
        self.log.debug(arg)
        start = time()
        with open(log_file, 'w') as log_handle:
            try:
                subprocess.run(arg, shell=True, check=True, stdout=log_handle, stderr=log_handle)
            except (OSError, subprocess.CalledProcessError) as e:
                if not no_exit:
                    self.log.exception('returned: ' + str(e.returncode)
                                       + e.output.decode('utf-8') if e.output is not None else '')
                    sys.exit(1)
                else:
                    raise e
        self.log.debug('%.2f seconds for %s' % (time() - start, info))



# 2021 Leo Kaindl

import logging
from pathlib import Path

import gi
from Bio import SeqIO

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from ab12phylo.gtk_base import ab12phylo_app_base
from ab12phylo.gtk_proj import picklable_liststore

LOG = logging.getLogger(__name__)
PAGE = 0


class io_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        # blink the progress bar so the window allocates a size
        iface.prog_bar.props.visible = True
        iface.win.show_all()
        iface.prog_bar.props.visible = False

        # trace file types
        # create a TreeView model
        iface.file_type_model = picklable_liststore(str, bool)
        [iface.file_type_model.append([file_type, False]) for
         file_type in ['.ab1', '.seq', '.fasta', '.fa', '.txt']]

        # check ABI traces by default
        iface.file_type_model[0][1] = True

        iface.view_filetypes.set_model(iface.file_type_model)
        iface.view_filetypes.set_headers_visible(False)
        iface.view_filetypes.append_column(
            Gtk.TreeViewColumn(title='Filetype', cell_renderer=Gtk.CellRendererText(), text=0))
        crt = Gtk.CellRendererToggle()
        crt.connect('toggled', lambda widget, path: iface.file_type_model.set(
            iface.file_type_model.get_iter(path), [1], [not iface.file_type_model[path][1]]))
        iface.view_filetypes.append_column(
            Gtk.TreeViewColumn(title='Selected', cell_renderer=crt, active=1))

        # trace files
        iface.add_refs.connect('clicked', self.add_manually, data.trace_store, 'references')

        # initial or old number of entries in model used for scrolling to end after adding
        iface.file_nums = dict()

        # set up the file paths tables
        for mo, tv, file_type, col in zip([data.trace_store, data.plate_store],
                                          [iface.view_trace_path, iface.view_csv_path],
                                          ['trace', 'csv'], [7, 3]):
            tv.set_model(mo)
            tv.set_headers_visible(False)
            tv.append_column(Gtk.TreeViewColumn(title='Paths',
                                                cell_renderer=Gtk.CellRendererText(editable=True),
                                                foreground=col, text=0))
            tv.connect('size-allocate', self.scroll_to_end, tv, mo)
            iface.file_nums[mo] = 0
            sel = tv.get_selection()
            sel.set_mode(Gtk.SelectionMode.MULTIPLE)

            iface.__getattribute__('add_%s_folder' % file_type) \
                .connect('clicked', self.add_folder, file_type, mo)
            iface.__getattribute__('add_%s_manual' % file_type) \
                .connect('clicked', self.add_manually, mo)
            try:
                iface.__getattribute__('add_%s_whitelist' % file_type) \
                    .connect('clicked', self.add_manually, mo, 'whitelists')
            except AttributeError:
                pass  # wellsplate whitelist not planned

            iface.__getattribute__('remove_%s' % file_type) \
                .connect('clicked', self.delete_files_from_input_selection, PAGE, sel)
            tv.connect('key-press-event', self.tv_keypress, PAGE, sel)
            iface.__getattribute__('delete_all_%s' % file_type) \
                .connect('clicked', self.delete_files_from_input_selection, PAGE, sel, mo)

        # horizontal scaling
        iface.zoomer.adj = iface.qal_scale.get_adjustment()
        iface.zoomer.adj.configure(1, .2, 2.6, .2, 0, 0)
        iface.qal_scale.set_digits(1)
        iface.gbl_scale.set_digits(1)
        iface.tree_scale.set_digits(1)
        iface.qal_scale.add_mark(1, Gtk.PositionType.BOTTOM, None)
        iface.gbl_scale.add_mark(1, Gtk.PositionType.BOTTOM, None)
        iface.tree_scale.add_mark(1, Gtk.PositionType.BOTTOM, None)
        iface.zoomer.bak = iface.zoomer.adj.get_value()
        iface.zoomer.handle = iface.zoomer.adj.connect_after('value-changed', self.x_scale, iface.zoomer)
        iface.zoomer.sizes = dict()  # will save with the image parent name here

    def add_folder(self, widget, file_type, model):
        if file_type == 'trace':
            file_types = {a[0] for a in self.iface.
                file_type_model.get_column((0, 1)) if a[1]}
        else:
            file_types = {'.csv'}

        dialog = Gtk.FileChooserDialog(title='select folder(s)', parent=None, select_multiple=True,
                                       action=Gtk.FileChooserAction.SELECT_FOLDER)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        duplicates, new_paths = list(), list()
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            for folder in dialog.get_filenames():
                try:
                    folder = Path(folder).resolve()
                    # get all the matching files
                    new_paths = [list(folder.glob('**/*%s' % ext)) for ext in file_types]
                    # flatten the LoL
                    new_paths = [item for sublist in new_paths for item in sublist]
                    # where to write
                    kw = 'csv' if '.csv' in file_types else 'trace'
                    # append to the ListStore
                    model, new_paths, duplicates = self.add_new_entries(model, new_paths, kw)
                except UnicodeDecodeError as ex:
                    LOG.info(ex)
            self.refresh()
            LOG.debug('found %d new paths in folder(s)' % len(new_paths))
        dialog.destroy()

        if duplicates:
            self.show_notification('Files already selected:', duplicates)

    def add_manually(self, widget, model, *args):
        """
        Load trace, wellsplate or reference paths into the appropriate GtkListStore,
        also remembering if this is a reference.
        """
        args = ['nope'] if not args else args
        dialog = Gtk.FileChooserDialog(title='select files', parent=None,
                                       action=Gtk.FileChooserAction.OPEN, select_multiple=True)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        not_found, duplicates = list(), list()
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            # read the filenames
            filenames = dialog.get_filenames()
            if 'whitelists' in args:
                # read lines from the files
                extracted = list()
                for whitelist in filenames:
                    try:
                        extracted += open(whitelist, 'r').read().strip().split('\n')
                    except UnicodeDecodeError as ex:
                        LOG.info(ex)
                filenames = extracted

            # sort extracted into valid and invalid file paths
            new_paths = list()
            for string_path in filenames:
                path = Path(string_path).resolve()
                if path.is_file():
                    new_paths.append(path)
                else:
                    not_found.append(string_path)

            # where to write
            if widget in [self.iface.add_csv_folder, self.iface.add_csv_manual]:
                kw = 'csv'
            elif widget == self.iface.add_refs:
                kw = 'ref'
            else:
                kw = 'trace'
            # append to ListStore
            model, new_paths, duplicates = self.add_new_entries(model, new_paths, kw)
            self.refresh()
            LOG.info('added %d paths' % len(new_paths))
        dialog.destroy()

        if not_found or duplicates:
            self.show_notification('File troubles', ['not found:%s' % f for f in not_found] +
                                   ['duplicate:%s' % f for f in duplicates], secs=0)

    def add_new_entries(self, model, new_paths, *args):
        """
        Adds new entries to a list and returns the modified list
        as well as lists of both newly added entries and observed duplicates.
        If *args, the new entries are reference sequences and the value in
        the color column will change accordingly
        :param model: the ListStore
        :param new_paths: obvs
        :param iface: the namespace containing all named widgets of a ab12phylo_app_base object
        :return:
        """
        if 'ref' in args:
            color = self.iface.AQUA
            is_ref = True
        elif 'trace' in args or 'csv' in args:
            color = None  # iface.FG
            is_ref = False
        else:
            raise RuntimeWarning(f'weird file type, neither ref, trace or csv')

        old_entries = [line[0] for line in model]
        dups, news = list(), list()

        for ppath in new_paths:
            path = str(ppath)
            if path in old_entries:
                dups.append(path)
            else:
                if 'csv' in args:
                    model.append([path, ppath.name, '', color])
                elif ppath.suffix.lower() != '.ab1' and not is_ref:
                    # no ABI trace, no ref and several records in the file -> list separately
                    rids = [r.id for r in SeqIO.parse(ppath, 'fasta')]
                    if len(rids) > 1:
                        [model.append([path + '~~' + rid, rid, '', '', '',
                                       False, False, color]) for rid in rids]
                    else:
                        model.append([path, ppath.name, '', '', '', is_ref, False, color])
                else:  # ABI trace file
                    model.append([path, ppath.name, '', '', '', is_ref, False, color])
                news.append(path)
        if len(news) > 0:
            self.set_changed(PAGE)
        return model, news, dups

    def scroll_to_end(self, widget, rectangle, tv, mo):
        """After new entries have been added to it, the TreeView will scroll to its end"""
        if self.iface.notebook.get_current_page() != 0:
            return
        if mo not in self.iface.file_nums or len(mo) > self.iface.file_nums[mo]:
            adj = tv.get_vadjustment()
            adj.set_value(adj.get_upper() - adj.get_page_size())
        self.iface.file_nums[mo] = len(mo)

    def refresh(self):
        """
        Adjusts the display of the number of selected trace files on the files page
        and toggles reading wellsplates or not, inactivating respective GUI elements.
        :return:
        """
        num_traces = len(self.data.trace_store)
        if num_traces > 0:
            self.iface.trace_number.set_label('%d files' % num_traces)
            self.iface.trace_number.set_visible(True)
        else:
            self.iface.trace_number.set_label('0 files')
            self.iface.trace_number.set_visible(False)



# 2021 Leo Kaindl

import logging
import threading
from pathlib import Path
from time import sleep

import gi
import matplotlib.pyplot as plt
import numpy as np
from Bio import SeqIO
from matplotlib.backends.backend_gtk3agg import FigureCanvasGTK3Agg as FigureCanvas
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ab12phylo import repo
from ab12phylo_cmd.filter import trim_ends, mark_bad_stretches
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
plt.set_loglevel('warning')
PAGE = 2


class qal_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.accept_rev.set_active(data.qal.accept_rev)
        iface.rev_handler = iface.accept_rev.connect('clicked', self.parse, None)
        iface.accept_nophred.set_active(True)
        iface.accept_nophred.connect('clicked', self.parse, None)

        iface.min_phred.set_adjustment(Gtk.Adjustment(value=30, upper=61, lower=0,
                                                      step_increment=1, page_increment=1))
        iface.min_phred.set_numeric(True)
        iface.min_phred.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)

        for w_name in ['min_phred', 'trim_out', 'trim_of', 'bad_stretch']:
            wi = iface.__getattribute__(w_name)
            data.qal.__setattr__(w_name, int(wi.get_text()))
            wi.connect('changed', self.edit_numerical_entry)
            wi.connect('focus_out_event', self.parse)
        iface.min_phred.disconnect_by_func(self.parse)  # MARK leave this line alone

        for w_name in ['accept_rev', 'accept_nophred']:
            data.qal.__setattr__(w_name, iface.__getattribute__(w_name).get_active())

        for wi in [iface.trim_out, iface.trim_of, iface.bad_stretch]:
            wi.connect('key-press-event', self.edit_numerical_entry_up_down)

        # init row annotation
        iface.view_qal.set_model(data.qal_model)
        iface.view_qal.set_headers_visible(False)
        iface.view_qal.append_column(Gtk.TreeViewColumn(
            title='id', cell_renderer=Gtk.CellRendererText(),
            text=0, underline=2, strikethrough=3))
        iface.view_qal.set_tooltip_column(1)
        # crt = Gtk.CellRendererToggle(radio=False)
        # crt.props.indicator_size = 13
        # iface.view_qal.append_column(Gtk.TreeViewColumn(
        #     title='no phreds', active=1, cell_renderer=crt))

        sel = iface.view_qal.get_selection()
        sel.set_mode(Gtk.SelectionMode.MULTIPLE)
        sel.connect('changed', self.keep_visible, iface.view_qal,
                    iface.parallel_qal.props.vadjustment.props, iface.tempspace)
        iface.view_qal.connect(
            'check-resize', self.get_height_resize, iface.qal_spacer, [iface.qal_win])
        # in-preview deletion
        iface.view_qal.connect(
            'key_press_event', self.delete_and_ignore_rows, PAGE, sel)
        iface.qal_eventbox.connect_after(
            'button_press_event', self.select_seqs, PAGE, iface.zoomer,
            iface.view_qal, iface.tempspace)  # in-preview selection
        iface.qal_win.connect('scroll-event', self.xy_scale, PAGE)  # zooming

        # connect gene switcher
        iface.gene_handler = iface.gene_roll.connect('changed', self.select_gene_and_redo)

    def init_gene_roll(self):
        """
        Initialize gene switcher combo box with the
        previously selected gene from the dataset.
        """
        data = self.data
        iface = self.iface
        with iface.gene_roll.handler_block(iface.gene_handler):
            iface.gene_roll.remove_all()
            genes = list(data.genes)
            if not genes:
                return
            [iface.gene_roll.append_text(gene) for gene in genes]
            if len(genes) > 1:
                iface.gene_roll.insert_text(0, 'all')
                genes.insert(0, 'all')
            if data.gene_for_preview and data.gene_for_preview in genes:
                idx = genes.index(data.gene_for_preview)
            else:
                idx = 0
                data.gene_for_preview = 'all'
            iface.gene_roll.set_active(idx)

    def select_gene_and_redo(self, *args):
        """
        A page-independent handler for selecting a different gene.
        Currently, iface.gene_roll is only visible from one page, so a bit useless.
        :param args:
        :return:
        """
        page = self.iface.notebook.get_current_page()
        if page == 2:  # trim preview
            self.parse(self.iface.gene_roll, None)
            self.start_trim()

    def refresh(self):
        data = self.data
        iface = self.iface
        with iface.accept_rev.handler_block(iface.rev_handler):
            iface.accept_rev.set_active(data.search_rev and iface.accept_rev.get_active())
            iface.accept_rev.set_sensitive(data.search_rev)

        if not (self.wd / repo.PATHS.preview).exists() or 0 in data.qal_shape:
            self.start_trim()
            return

        # place the png preview
        self.load_image(iface.zoomer, PAGE, iface.qal_eventbox, self.wd / repo.PATHS.preview,
                        data.qal_shape[0] * self.get_hadj(), data.qal_shape[1])
        self.load_colorbar(iface.palplot)
        self.win.show_all()

    def reload_ui_state(self):
        for w_name in ['min_phred', 'trim_out', 'trim_of', 'bad_stretch']:
            self.iface.__getattribute__(w_name).set_text(str(self.data.qal.__getattribute__(w_name)))
        for w_name in ['accept_rev', 'accept_nophred']:
            self.iface.__getattribute__(w_name).set_active(self.data.qal.__getattribute__(w_name))

    def parse(self, widget, event):
        """
        Parse the content of a widget. Hitting enter in an entry still calls this, which is good.
        :param widget: The element to parse and inspect for changes
        :param event: passed by some signals -> focus_out_event
        :return:
        """
        data = self.data
        iface = self.iface
        LOG.debug('parsing for re-draw')
        w_name = widget.get_name()
        pre = None
        try:
            pre = data.qal.__getattribute__(w_name)
        except (KeyError, AttributeError):
            LOG.error('%s not in params' % w_name)
        # getting new value depends on widget type
        if widget == iface.gene_roll:
            now = widget.get_active_text()
        elif widget in [iface.accept_rev, iface.accept_nophred]:
            now = widget.get_active()
        else:
            try:
                now = int(widget.get_text())
            except ValueError:
                now = 0
        self.delete_event(widget, event)

        # cause re-drawing if something changed
        if pre == now:
            LOG.debug('no change at parsing')
            return
        data.qal.__setattr__(w_name, now)
        if data.qal.trim_out > data.qal.trim_of:
            self.show_notification('cannot draw %d from %d' %
                                   (data.qal.trim_out, data.qal.trim_of))
            widget.set_text('0')
        else:
            self.set_changed(PAGE)

    def start_trim(self):
        """
        For non-empty gene set, start a background re-drawing thread
        and return to the main loop.
        :return:
        """
        data = self.data
        iface = self.iface
        # called after files are read
        if not data.genes or iface.thread.is_alive() \
                or iface.notebook.get_current_page() != PAGE:
            self.show_notification('Busy', secs=1)
            return

        if not data.seqdata:
            self.start_read(run_after=[self.start_trim])
            return

        data.gene_ids = {g: set(gd.keys()) for g, gd in data.seqdata.items() if g in data.genes}
        data.qal.accept_rev = iface.accept_rev.get_active()
        self.parse(iface.min_phred, None)  # annoying SpinButton

        LOG.debug('start-up redraw')
        data.qal_model.clear()
        sleep(.1)
        iface.thread = threading.Thread(target=self.do_trim)
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        # return to main loop

    def do_trim(self):
        """
        Iterate over records and trim, create a matrix representation
        of the valid characters and plot it.
        :return:
        """
        data = self.data
        iface = self.iface
        # parameters are up-to-date
        LOG.debug('re-draw trim preview')
        p = data.qal
        rows = list()
        data.gene_for_preview = p.gene_roll
        genes_for_preview = data.genes if p.gene_roll == 'all' else {p.gene_roll}
        ignore_ids = data.qal.ignore_ids if 'ignore_ids' in data.qal \
            else {g: set() for g in data.genes}  # regulated at delete_and_ignore_rows
        iface.text = 'creating matrix'
        LOG.debug(iface.text)
        iface.i = 0
        iface.k = sum([len(data.seqdata[gene]) for gene in genes_for_preview
                       if gene != 'no match']) + 4  # number of all records + extra
        shared_ids = set.intersection(*data.gene_ids.values())

        if not shared_ids:
            msg = 'No shared sequences between all genes (%s)' % ','.join(data.genes)
            self.show_notification(msg)
            LOG.warning(msg)
            sleep(.1)
            GObject.idle_add(self.stop_trim, True)
            return True

        try:
            for rid, gene in data.record_order:
                # skip records from other genes for the trimming preview
                if gene not in genes_for_preview or \
                        gene in ignore_ids and rid in ignore_ids[gene]:
                    continue

                iface.i += 1
                # maybe skip reversed seqs
                if not p.accept_rev and data.metadata[gene][rid]['is_rev']:
                    continue

                record = data.seqdata[gene][rid]
                try:
                    record = trim_ends(record, p.min_phred, (p.trim_out, p.trim_of), trim_preview=True)
                    record = mark_bad_stretches(record, p.min_phred, p.bad_stretch)
                    has_qal, is_bad = True, False
                    row = repo.seqtoint(record)
                except AttributeError:
                    # accept references anyway, but maybe skip no-phred ones
                    is_ref = 'accession' in data.metadata[gene][rid]
                    if not is_ref and not p.accept_nophred:
                        continue
                    has_qal, is_bad = is_ref, False
                    row = repo.seqtoint(record)
                except ValueError:
                    has_qal, is_bad = True, True
                    row = repo.seqtogray(record)
                rows.append(row)
                underline = not has_qal if rid in shared_ids else 4
                data.qal_model.append([rid, gene, underline, is_bad])

        except KeyError as ke:
            LOG.exception(ke)

        if not rows:
            msg = 'No sequence data remains'
            self.show_notification(msg)
            LOG.warning(msg)
            sleep(.1)
            GObject.idle_add(self.stop_trim)
            return True

        iface.text = 'tabularize'
        LOG.debug(iface.text)
        max_len = max(map(len, rows))
        array = np.array([row + repo.seqtoint(' ') * (max_len - len(row)) for row in rows])
        # make gaps transparent
        array = np.ma.masked_where(array > repo.toint('else'), array)
        iface.i += 1
        iface.text = 'plot'
        LOG.debug(iface.text)

        # adjust maximum size
        scale = 6
        while max(array.shape) * scale > 2 ** 14:
            scale -= 1
        LOG.debug('scaling qal with %d' % scale)

        f = Figure()
        f.set_facecolor('none')
        f.set_figheight(array.shape[0] / repo.DPI)
        f.set_figwidth(array.shape[1] / repo.DPI)
        ax = f.add_subplot(111)
        ax.axis('off')
        f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)
        mat = ax.matshow(array, alpha=1, cmap=ListedColormap(repo.colors),
                         vmin=-.5, vmax=len(repo.colors) - .5, aspect='auto')

        iface.i += 1
        iface.text = 'save PNG'
        LOG.debug(iface.text)
        Path.mkdir(self.wd / repo.PATHS.preview.parent, exist_ok=True)
        f.savefig(self.wd / repo.PATHS.preview, transparent=True,
                  dpi=scale * repo.DPI, bbox_inches='tight', pad_inches=0.00001)
        plt.close(f)

        data.qal_shape[0] = array.shape[1]
        data.qal_shape[1] = self.get_height_resize(
            iface.view_qal, None, iface.qal_spacer, [iface.qal_win])

        if iface.rasterize.props.active:
            iface.text = 'place PNG'
            LOG.debug(iface.text)
            sleep(.05)
            self.load_image(iface.zoomer, PAGE, iface.qal_eventbox, self.wd / repo.PATHS.preview,
                            data.qal_shape[0] * self.get_hadj(), data.qal_shape[1])
        else:
            iface.text = 'place vector'
            LOG.debug(iface.text)
            canvas = FigureCanvas(f)  # a Gtk.DrawingArea
            canvas.set_size_request(data.qal_shape[0] * self.get_hadj(), data.qal_shape[1])
            try:
                ch = iface.qal_eventbox.get_child()
                if ch:
                    iface.qal_eventbox.remove(ch)
                iface.qal_eventbox.add(canvas)
            except Exception as ex:
                LOG.error(ex)
        iface.i += 1
        self.load_colorbar(iface.palplot)

        iface.text = 'idle'
        iface.frac = 1
        sleep(.1)
        GObject.idle_add(self.stop_trim)
        return True

    def stop_trim(self, errors=False):
        self.iface.thread.join()
        self.win.show_all()
        self.set_errors(PAGE, errors)
        self.iface.prog_bar.set_text('idle')
        LOG.info('qal thread idle')
        return False

    def trim_all(self, run_after=None):
        """
        Trim all SeqRecords in project_dataset.seqdata to sequence strings
        and write two collated .fasta files per gene, one only containing
        the records shared across all genes. If necessary, trace files are
        re-read before and will be deleted from memory afterwards for a
        smaller project file. Plot and place png previews of trimming result.
        :param run_after: [functions(gui)] when finished, usually flip page.
        :return:
        """
        data = self.data
        iface = self.iface

        if not data.seqdata:
            LOG.debug('re-reading files')
            self.start_read(run_after=[self.trim_all])
            return

        p = data.qal
        ignore_ids = p.ignore_ids if 'ignore_ids' in p \
            else {g: set() for g in data.genes}  # regulated at delete_and_ignore_rows
        LOG.debug('trim and filter all sequences')
        for gene, genedata in data.seqdata.items():
            Path.mkdir(self.wd / gene, exist_ok=True)
            genemeta = data.metadata[gene]

            # do actual trimming
            for _id in data.gene_ids[gene]:
                record = genedata.pop(_id)
                record_meta = genemeta[_id]

                if gene in ignore_ids and _id in ignore_ids[gene]:
                    record_meta['quality'] = 'manually dropped at trim_data'
                    continue
                if not p.accept_rev and record_meta['is_rev']:
                    record_meta['quality'] = 'disallowed reverse reads'
                    continue
                try:
                    record = trim_ends(record, p.min_phred, (p.trim_out, p.trim_of))
                    record = mark_bad_stretches(record, p.min_phred, p.bad_stretch)
                except ValueError:
                    record_meta['quality'] = 'low quality'
                    continue
                except AttributeError:
                    is_ref = 'accession' in record_meta
                    if not p.accept_nophred and not is_ref:
                        continue
                    if not is_ref:
                        record_meta['quality'] = 'no phreds'
                # put back only the good records
                genedata[_id] = record

            # update dict of legal ids
            data.gene_ids[gene] = set(genedata.keys())

        # re-filter for shared entries
        shared_ids = set.intersection(*data.gene_ids.values())
        # re-loop seqdata
        LOG.debug('writing collated .fasta files')
        for gene, genedata in data.seqdata.items():
            # write to file
            with open(str(self.wd / gene / (gene + '_all.fasta')), 'w') as fasta:
                SeqIO.write(genedata.values(), fasta, 'fasta')
            # write only records shared across all genes to fasta for MSA
            with open(str(self.wd / gene / (gene + '.fasta')), 'w') as fasta:
                SeqIO.write([record for _id, record in genedata.items()
                             if _id in shared_ids], fasta, 'fasta')

            genemeta = data.metadata[gene]
            for _id in {_id for _id in genedata.keys() if _id not in shared_ids}:
                if 'quality' not in genemeta[_id]:
                    genemeta[_id]['quality'] = 'not in all genes'

        LOG.debug('writing metadata, deleting seqdata')
        self.write_metadata()
        data.seqdata.clear()

        self.set_changed(PAGE, False)
        self.set_changed(PAGE + 1, True)
        if run_after:
            [run() for run in run_after]
        return

    def delete_event(self, widget, event):
        return False



# 2020 Leo Kaindl

"""
The AB12PHYLO command line interface defines possible options and parses valid arguments
from user input via :class:`sys.argv`, supplemented by the :file:`config/config.yaml` file.
Arguments will be saved as an :class:`argparse.Namespace` object and directly accessed by the
:module:`main` module. Additionally, this module initiates logging.
"""

import argparse
import configparser
import logging
import os
import random
import shutil
import sys
import zipfile
from os import path
from pathlib import Path

import requests
import yaml

from ab12phylo_cmd import phylo
from ab12phylo_cmd.filter import fetch_non_python_tools
from ab12phylo_cmd.__init__ import __version__, __date__


class parser(argparse.ArgumentParser):
    """
    The command line parser of the package.
    Run `ab12phylo-cmd -h` to display all available options.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(prog='ab12phylo-cmd')

        # if empty commandline, show info and help:
        if not args[0]:
            print('AB12PHYLO commandline version %s built on %s\n' % (__version__, __date__))
            args = (['-h'],) + args[1:]

        # [modes]
        mod = parser.add_argument_group(self, title='RUN MODES')
        parts = mod.add_mutually_exclusive_group()
        parts.add_argument('-p1', '--prepare', action='store_true',
                           help='run first part of ab12phylo-cmd, including BLAST but excluding RAxML-NG/IQ-Tree.')
        parts.add_argument('-p2', '--finish', action='store_true',
                           help='run second part of ab12phylo-cmd, beginning with RAxML-NG/IQ-Tree.')
        parts.add_argument('-px', '--add_xml', action='store_true',
                           help='after -p1 run; only read BLAST results. Pass file via -xml.')
        mod.add_argument('-viz', '--visualize', action='store_true',
                         help='invoke ab12phylo-visualize by appending ab12phylo-cmd command.')
        mod.add_argument('-view', '--view', action='store_true',
                         help='invoke ab12phylo-view by appending ab12phylo-cmd command.')

        # [I/O]
        ion = parser.add_argument_group(self, title='FILE I/O')
        ion.add_argument('-dir', '--dir',
                         help='output directory. Defaults to \'./results\'')
        ion.add_argument('-g', '--genes', nargs='+',
                         help='gene(s) to be considered; first argument defines gene for species annotation. '
                              'If set, only ABI traces with a filename matching one of these patterns will be read.')
        ion.add_argument('-abi', '--abi_dir',
                         help='root directory of all ABI trace files. Defaults to current working directory',
                         type=lambda arg: arg if path.isdir(arg) else self.error(
                             '%s: invalid ABI trace file directory' % arg))
        ion.add_argument('-abiset', '--abi_set',
                         help='whitelist file defining subset of ABI traces for the analysis. '
                              'Files must be in or below provided \'--abi_dir\' directory.',
                         type=lambda arg: arg if path.isfile(arg) else self.error(
                             '%s: invalid whitelist file' % arg))
        ion.add_argument('-sampleset', '--sample_set',
                         help='whitelist file defining subset of sample IDs for the analysis. '
                              'Different versions of a sample will be included.',
                         type=lambda arg: arg if path.isfile(arg) else self.error(
                             '%s: invalid sample whitelist' % arg))
        ion.add_argument('-csv', '--csv_dir',
                         help='root directory of .csv files with well-to-isolate coordinates.',
                         type=lambda arg: arg if path.isdir(arg) else 'ignore')
        ion.add_argument('-r2', '--regex_csv', help='RegEx to parse the wellsplate number from a .csv filename. '
                                                    'Use a single capturing group, and double quotes in bash.')
        rxp = ion.add_mutually_exclusive_group()
        rxp.add_argument('-r1', '--regex_abi', help='RegEx to parse plate number, gene name and well position '
                                                    'from an .ab1 filename in that order. Use double quotes in bash.')
        rxp.add_argument('-r3', '--regex_3', nargs=3,
                         help='alternative to --regex_abi. 3 regular expressions to parse plate number, gene name and '
                              'well from an .ab1 filename. Provide the regular expressions in that order without commas'
                              ', but with double quotes from bash. Use a single capturing group in each regex.')
        ion.add_argument('-r4', '--regex_rev', help='RegEx to identify reverse reads from their .ab1 filename.')

        refs = ion.add_mutually_exclusive_group()
        refs.add_argument('-rf', '--ref', nargs='+',
                          help='optional paths of .fasta-formatted files containing reference sequences. '
                               'Files will be matched to genes by order if --genes is set, otherwise by filename.',
                          type=lambda arg: arg if path.isfile(arg) else self.error(
                              'invalid file path(s) with .fasta-formatted reference sequences:\n%s' % arg))
        refs.add_argument('-rd', '--ref_dir', type=self._valid_ref_dir,
                          help='directory of .fasta files with reference sequences. Files will be matched to genes '
                               'by their filename. Set only one option from {--ref, --ref_dir}.')

        # [quality]
        qal = parser.add_argument_group(self, title='QUALITY')
        qal.add_argument('-qal', '--min_phred', type=int,
                         help='minimal phred quality score to define \'good\' bases in ABI trace files.')
        qal.add_argument('-bad', '--bad_stretch', type=int,
                         help='number of consecutive \'bad bases\': Any sequence of bases in an ABI trace file '
                              'with a phred quality score below the minimum and at least as long as the number '
                              'supplied here will be replaced by a sequence of Ns of equal length.')
        qal.add_argument('-end', '--end_ratio', type=self._valid_end_ratio,
                         help='defines a \'good end\' of a sequence in an ABI trace file for trimming. '
                              'Enter as "<int>/<int>".')

        # [BLAST]
        bla = parser.add_argument_group(self, title='BLAST')
        skips = bla.add_mutually_exclusive_group()
        skips.add_argument('-local', '--no_remote', action='store_true',
                           help='NCBI BLAST API queries are de-prioritized very quickly. Set this flag to skip online '
                                'nucleotide BLAST for seqs missing from the local database.')
        skips.add_argument('-none', '--no_BLAST', action='store_true',
                           help='skip BLAST entirely.')
        skips.add_argument('-remote', '--no_local', action='store_true',
                           help='BLAST only in remote database; discouraged.')
        skips.add_argument('-xml', '--BLAST_xml', nargs='+', help='supply available BLAST results as .XML files.')
        bla.add_argument('-db', '--db', help='BLAST+ database to use. Will attempt to download or update this db from '
                                             'https://ftp.ncbi.nlm.nih.gov/blast/db/ unless provided via -dbpath.')
        bla.add_argument('-dbpath', '--dbpath',
                         help='path to directory with a BLAST+ database. Set this option for a user-created database '
                              'or if ab12phylo-cmd is not allowed FTP access. You might have to define -db as well.',
                         type=lambda arg: arg if path.isdir(arg) else self.error(
                             'invalid path to directory containing BLAST database'))
        bla.add_argument('-remotedb', '--remote_db',
                         help='NCBI database to search for sequences not found locally. Default is \'nt\' for DNA.')

        # [MSA]
        msa = parser.add_argument_group(self, title='MSA')
        msa.add_argument('-algo', '--msa_algo', choices=['clustalo', 'mafft', 'muscle', 't_coffee'],
                         help='select an algorithm to build the Multiple Sequence Alignment. Default is MAFFT.')
        msa.add_argument('-gbl', '--gblocks', choices=['skip', 'relaxed', 'balanced', 'default', 'strict'],
                         help='activate/set MSA trimming with Gblocks.')

        phy = parser.add_argument_group(self, title='ML TREE INFERENCE')
        phy.add_argument('-tool', '--ml_tool', choices=['raxml-ng', 'iqtree2'],
                         help='select a tool to re-construct a phylogenetic tree. RAxML-NG is the default '
                              'for Linux, but not available for Windows.')
        phy.add_argument('-st', '--start_trees', type=self._valid_start_trees,
                         help='numbers of starting trees for raxml-ng / iqtree2 tree inference: '
                              '[<int random trees>,<int parsimony-based trees>].')
        phy.add_argument('-bst', '--bootstrap', type=self._valid_bootstrap,
                         help='number of bootstrap trees. The MRE bootstrap convergence test in RAxML-NG may stop '
                              'iteration before the passed number is reached, but this is unlikely. See '
                              'https://doi.org/10.1089/cmb.2009.0179')
        phy.add_argument('-uf', '--ultrafast', action='store_true',
                         help='for tree inference with IQ-Tree 2, ultrafast bootstrapping can be enabled. '
                              'If a value lower than 1000 is passed to `--bootstrap`, the number of iterations will '
                              'automatically be set to this minimum accepted value.')
        mod = phy.add_mutually_exclusive_group()
        mod.add_argument('-evomodel', '--evomodel', help='evolutionary model for tree inference. The default '
                                                         'is GTR+I+G, and no checks are performed!')
        mod.add_argument('-findmodel', '--findmodel', action='store_true',
                         help='FOR IQ-TREE2 ONLY: infer an evolutionary model using ModelFinder before '
                              're-constructing a tree. https://doi.org/10.1038/nmeth.4285')
        phy.add_argument('-s', '--seed', type=int,
                         help='seed value for reproducible tree inference results. Will be random if not set.')
        phy.add_argument('-metric', '--metric', choices=['TBE', 'FBP'],
                         help='bootstrap support metric: Either Felsenstein Bootstrap Proportions (FBP) '
                              'or Transfer Bootstrap Expectation (TBE). https://doi.org/10.1038/s41586-018-0043-0')

        # [visualize]
        viz = parser.add_argument_group(self, title='VISUALIZATION')
        viz.add_argument('result_dir', nargs='?',
                         help='ONLY FOR AB12PHYLO-VISUALIZE/-VIEW: path to earlier run. Pass without keyword!')
        viz.add_argument('-msa_viz', '--msa_viz', nargs='*', choices=['pdf', 'png'],
                         help='also render a rectangular tree with MSA color matrix in chosen format(s). '
                              'Takes some extra time.')
        viz.add_argument('-threshold', '--threshold', type=float,
                         help='limit between 0 and 1 for support value color switch.')
        viz.add_argument('-out_fmt', '--out_fmt', nargs='+', choices=['pdf', 'png', 'svg'],
                         help='output file format for tree graphics.')
        viz.add_argument('-md', '--min_dist', type=float,
                         help='minimal phylogenetic distance between any pair of samples for RAxML-NG.')
        viz.add_argument('-mpd', '--min_plot_dist', type=float,
                         help='minimal artificial distance of a node in the visualization to its parent.')
        viz.add_argument('-drop', '--drop_nodes', nargs='+', type=int,
                         help='drop the node(s) with this idx and its descendants from the tree. '
                              'Use the motif subtree search to find the node or MRCA idx of the target.')
        viz.add_argument('-replace', '--replace_nodes', nargs='+', type=int,
                         help='replace the node(s) with this idx and its descendants with a placeholder.')
        viz.add_argument('-root', '--root', type=int, help='root the tree at the node with this index.')
        viz.add_argument('-supp', '--print_supports', action='store_true',
                         help='print the support values in percent of the optimal value in the rectangular tree.')

        # [popgen]
        gen = parser.add_argument_group(self, title='POPULATION GENETICS')
        gen.add_argument('-gap', '--gap_share', type=self._valid_threshold,
                         help='maximum share of gaps at an MSA site that will be ignored.')
        gen.add_argument('-unk', '--unknown_share', type=self._valid_threshold,
                         help='maximum acceptable proportion of unknown characters at an MSA site.')
        viz.add_argument('-poly', '--poly_allelic', action='store_true',
                         help='accept segregating sites with more than one mutant nucleotide.')

        # [misc]
        level = self.add_mutually_exclusive_group()
        level.add_argument('-i', '--info', action='store_true', help='show some more information in console output.')
        level.add_argument('-v', '--verbose', action='store_true', help='show all runtime information in console.')
        self.add_argument('-c', '--config',
                          default=path.abspath(path.join(path.dirname(__file__), 'config', 'config.yaml')),
                          type=lambda arg: arg if path.isfile(arg) else self.error('%s: invalid .config path'),
                          help='path to .yaml config file with defaults; command line arguments will override them.')
        self.add_argument('-nt', '--max_threads', type=int,
                          help='Limit the number of CPUs to use for AB12PHYLO.')
        self.add_argument('-version', '--version', action='store_true', help='print version information and exit.')
        self.add_argument('-test', '--test', action='store_true', help='Test run.')
        self.add_argument('-q', '--headless', action='store_true',
                          help='do not start a CGI server nor display in browser. For remote use.')
        self.add_argument('-init', '--initialize', action='store_true',
                          help='re-initialize ab12phylo-cmd: Search for existing BLAST+, '
                               'RAxML-NG and IQ-Tree installations, or re-run these.')

        self._initialize()

        self.args = self.parse_args(args[0])

        if self.args.version is True:
            sys.exit('ab12phylo: %s' % __version__)

        cfg = Path(__file__).resolve().parent / 'config' / 'conf.cfg'
        if self.args.initialize is True:
            if cfg.is_file():
                cfg.unlink()
            self._initialize()
            exit(0)

        # test: switch config + set verbose
        if self.args.test is True:
            print('--TEST RUN--', file=sys.stderr)
            self.args.config = path.abspath(path.join(path.dirname(__file__), 'config', 'test_config.yaml'))
            self.args.verbose = True

        # load additional info from config
        assert self.args.config is not None
        config = yaml.safe_load(open(self.args.config, 'r'))

        # if refs were set manually; and genes as well, or there is only 1 reference -> match refs to genes by order
        by_order = True if self.args.ref is not None \
                           and (self.args.genes is not None or len(self.args.ref) == 1) else False

        config_only = dict()
        # provide defaults for unset options
        for key, val in config.items():
            if key not in self.args:
                # for items without CLI equivalent (avoid bloated CLI)
                config_only[key] = val
            elif self.args.__dict__.get(key) in [None, False]:
                # access the namespace itself without var names -> access dict
                if key in ['abi_dir', 'csv_dir', 'blastdb', 'abi_set', 'sample_set', 'dir'] and val[0] == '$':
                    # deal with relative paths in config for test case
                    val = path.join(path.dirname(path.dirname(__file__)), val[1:])
                elif key == 'ref':
                    # split into list
                    val = [ref.strip() for ref in val.split(',')]
                    # make absolute paths
                    val = [path.join(path.dirname(path.dirname(__file__)), ref[1:])
                           if ref[0] == '$' else ref for ref in val]
                elif key == 'regex_abi' and self.args.regex_3 or key == 'regex_3' and self.args.regex_abi:
                    # do not interfere with user-defined exclusive group
                    continue
                elif key == 'findmodel' and self.args.evomodel or key == 'evomodel' and self.args.findmodel:
                    # as above
                    continue

                self.args.__dict__[key] = val

        # ab12phylo-cmd with --visualize or --view: guess real results path and skip re-parsing
        if len(kwargs) > 0 or self.args.visualize or self.args.view:
            if self.args.visualize:
                print('starting -viz re-plotting run', file=sys.stderr)
            elif self.args.view:
                print('re-view results', file=sys.stderr)
            # look in current working directory and ./results
            found = False
            for outer in [self.args.result_dir, self.args.dir, os.getcwd()]:
                if found:
                    break
                if outer is None:
                    continue
                elif outer == '.':
                    outer = os.getcwd()
                for inner in ['', 'results']:
                    if path.isfile(path.join(outer, inner, 'tree_TBE.nwk')):
                        # filename is hardcoded manually from below.
                        self.args.dir = path.join(outer, inner)
                        found = True
                        break
            if not found:
                sys.exit('Result files not found')

        else:  # normal case

            # move ref file list from ref_dir to ref
            if self.args.ref_dir:
                if type(self.args.ref_dir) == list:
                    self.args.ref = self.args.ref_dir
                else:
                    self.args.ref = self._valid_ref_dir(self.args.ref_dir)
                del self.args.ref_dir

            # check for duplicates in genes and references
            if self.args.genes is not None:
                if len(set(self.args.genes)) < len(self.args.genes):
                    self.error('duplicates in supplied genes')
            if self.args.ref is not None and len(set(self.args.ref)) < len(list(self.args.ref)):
                self.error('duplicates in supplied references')

            # now rebuild a command line and parse it again
            commandline = list()
            for key, val in self.args.__dict__.items():
                if key in ['genes', 'ref', 'regex_3', 'BLAST_xml', 'out_fmt', 'msa_viz'] and val is not None:
                    commandline.append('--%s' % key)
                    [commandline.append(v) for v in val]
                    if key == 'msa_viz' and not val:
                        commandline.append('png')
                elif val not in [None, False, True]:
                    commandline += ['--%s' % key, str(val)]
                elif val is True:
                    commandline.append('--%s' % key)

            self.args = self.parse_args(commandline)

            if sys.platform in ['win32', 'cygwin'] and self.args.ml_tool == 'raxml-ng' and not \
                    (self.args.visualize or self.args.view or self.args.prepare or self.args.add_xml):
                print(f'\033[91mWARNING: RAxML-NG was selected for ML inference on a Windows '
                      f'or cygwin system. This will likely fail.\033[0m Select IQ-Tree 2 by '
                      f'passing \'--ml_tool iqtree2\' or editing the equivalent line in '
                      f'{self.args.config}')
                if not self.args.headless:
                    answer = ''
                    while answer not in {'y', 'yes', 'n', 'no'}:
                        answer = input(f'Continue anyway? [y/n]').lower().strip()
                    if answer in {'n', 'no'}:
                        exit(1)

            # remember type of original ref option
            self.args.by_order = by_order

            # create output directory already
            if self.args.dir is not None:
                os.makedirs(self.args.dir, exist_ok=True)
            else:
                # write results to current directory.
                self.args.dir = ''

            # define a random seed if None was given
            if self.args.seed is None:
                self.args.seed = random.randint(0, 1000)

        # fetch the paths from the config
        cfg_parser = configparser.ConfigParser()
        cfg_parser.read(cfg)
        self.args.cfg = dict()
        if 'Paths' in cfg_parser:
            self.args.cfg.update(dict(cfg_parser['Paths']))

        # set some default values where options would be useless
        self.args.xml = path.join(self.args.dir, 'BLAST', 'local_blast+_result.xml')
        self.args.www_xml = path.join(self.args.dir, 'BLAST', 'online_blast_result.xml')
        self.args.bad_seqs = path.join(self.args.dir, 'bad_seqs.tsv')
        self.args.missing_samples = path.join(self.args.dir, 'missing_samples.tsv')
        self.args.tsv = path.join(self.args.dir, 'metadata.tsv')
        self.args.msa = path.join(self.args.dir, 'msa.fasta')
        self.args.new_msa = path.join(self.args.dir, 'msa_annotated.fasta')
        self.args.mview_msa = path.join(self.args.dir, 'msa_mview.html')
        self.args.topo = path.join(self.args.dir, 'topology_preview.png')
        self.args.missing_fasta = path.join(self.args.dir, 'missing.fasta')
        self.args.final_tree = path.join(self.args.dir, 'tree')
        self.args.annotated_tree = path.join(self.args.dir, 'tree_%s_annotated.nwk' % self.args.metric)
        if self.args.prepare:
            self.args.log = path.join(self.args.dir, 'ab12phylo-p1.log')
        elif self.args.finish:
            self.args.log = path.join(self.args.dir, 'ab12phylo-p2.log')
        elif self.args.add_xml:
            self.args.log = path.join(self.args.dir, 'ab12phylo-px.log')
        else:
            self.args.log = path.join(self.args.dir, 'ab12phylo.log')
        self.args.sep = 'SSSSSSSSSS'  # to visually separate genes in the concat MSA
        # now also load config-only defaults
        self.args.__dict__.update(config_only)

        # switching to visualize and view
        if self.args.visualize:
            self._init_log(self.args.log[:-4] + '-viz.log')
            log = logging.getLogger(__name__)
            log.debug('--AB12PHYLO-VISUALIZE--')
            log.debug(' '.join(args[0]))
            # copy config
            shutil.copy(src=self.args.config, dst=path.join(self.args.dir, 'used_config.yaml'))
            phylo.tree_build(self.args)
            sys.exit(0)

        if self.args.view:
            self.args.headless = False
            self._init_log(self.args.log[:-4] + '-view.log')
            log = logging.getLogger(__name__)
            log.debug('--AB12PHYLO-VIEW--')
            log.debug(' '.join(args[0]))
            # copy config
            shutil.copy(src=self.args.config, dst=path.join(self.args.dir, 'used_config.yaml'))
            phylo.tree_view(self.args.dir)
            sys.exit(0)

        # configure logging:
        if len(kwargs) > 0:
            self._init_log(self.args.log[:-4] + '-view-viz.log')
            log = logging.getLogger(__name__)
            log.debug(' '.join(args[0]))
        else:
            self._init_log(self.args.log)
            log = logging.getLogger(__name__)
            log.debug('--ARGS-- %s' % ' '.join(args[0]))
            log.debug('running AB12PHYLO v%s' % __version__)
            log.info('seed for this run: %s' % self.args.seed)
            if by_order is True:
                log.info('will match references to genes by order')

        # copy config
        shutil.copy(src=self.args.config, dst=path.join(self.args.dir, 'used_config.yaml'))

    def _initialize(self):
        """
        Check if ab12phylo-cmd has been run before, i.e. whether the :file:`conf.cfg`
        file is present. If not, search for BLAST+, RAxML-NG and iqtree2 installations
        via shutil. If a tool is not found on the $PATH or outdated, run prompts and
        installations. If the file is not present, always try downloading test data.
        :return:
        """
        _dir = Path(__file__).resolve().parent
        cfg = _dir / 'config' / 'conf.cfg'
        other_cfg = _dir.parent / 'ab12phylo' / 'conf.cfg'
        if cfg.is_file():
            return

        # init temporary console logger
        log = logging.getLogger(__name__)
        log.setLevel(logging.DEBUG)
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
        log.addHandler(sh)

        log.info('Initializing ab12phylo-cmd:')

        log.info('Downloading test data ...')
        r = requests.get('https://github.com/lkndl/ab12phylo/wiki/cmd_test_data.zip',
                         stream=True, timeout=12)
        zf = _dir / 'test_data.zip'
        with open(zf, 'wb') as file:
            for chunk in r.iter_content(chunk_size=128):
                file.write(chunk)
        with zipfile.ZipFile(zf, 'r') as zo:
            zo.extractall(zf.with_suffix(''))

        fetch_non_python_tools('-cmd', cfg, other_cfg, _dir / 'tools', log)
        sh.close()
        log.removeHandler(sh)

    def _valid_ref_dir(self, ref_dir):
        """
        Looks for .fasta files in --ref_dir arg, return as list. Trick :later move to --ref flag.
        :return: .fasta reference files as list
        """
        if not path.isdir(ref_dir):
            raise self.error('invalid references directory: %s' % ref_dir)

        ref_files = list()
        for root, dirs, files in os.walk(ref_dir):
            ref_files += [path.join(root, file) for file in files if file.endswith('.fasta')]

        if len(ref_files) == 0:
            raise self.error('no .fasta references found in directory: %s' % ref_dir)
        return ref_files

    def _valid_end_ratio(self, end_ratio):
        """Checks if --end_ratio argument for trimming is in right format and meaningful"""
        try:
            ratio = [int(d) for d in end_ratio.strip().split('/')]
            if len(ratio) == 2 and ratio[0] <= ratio[1]:
                return ratio
            else:
                raise ValueError
        except ValueError:
            raise self.error('invalid end ratio defined: %s' % end_ratio)

    def _valid_start_trees(self, start_trees):
        """Checks if --start_trees argument is in the right format: [int,int]"""
        try:
            start = [int(d) for d in start_trees[1:-1].split(',')]
            if len(start) == 2:
                return start
            else:
                raise ValueError
        except ValueError:
            raise self.error('invalid start trees: %s' % start_trees)

    def _valid_bootstrap(self, bootstrap):
        """Checks if --bootstrap argument is a number > 1"""
        try:
            bootstrap = int(bootstrap)
            if bootstrap > 1:
                return bootstrap
            else:
                raise ValueError
        except ValueError:
            raise self.error('Number of bootstrap trees must be an int > 1')

    def _valid_threshold(self, t):
        """Checks if popgen thresholds are floats in [0:0.5]"""
        try:
            thresh = float(t)
            if thresh < 0 or thresh > 1:
                raise ValueError
            else:
                return thresh
        except ValueError:
            raise self.error('invalid numerical float threshold: %s' % t)

    def _init_log(self, filename):
        """Initializes logging"""
        log = logging.getLogger()
        log.setLevel(logging.DEBUG)

        # init verbose logging to file
        fh = logging.FileHandler(filename=filename, mode='w')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter('%(asctime)s: %(levelname)s\t%(name)s\t%(message)s',
                                          datefmt='%Y-%m-%d %H:%M:%S'))
        log.addHandler(fh)

        # init shortened console logging
        sh = logging.StreamHandler(sys.stdout)
        if self.args.verbose:
            sh.setLevel(logging.DEBUG)
        elif self.args.info:
            sh.setLevel(logging.INFO)
        else:
            sh.setLevel(logging.WARNING)
        sh.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
        log.addHandler(sh)



# 2021 Leo Kaindl

import logging
import os
import shutil
import subprocess
import sys
import threading
from pathlib import Path
from time import sleep

import gi
import matplotlib.pyplot as plt
import numpy as np
from Bio import SeqIO
from matplotlib.backends.backend_gtk3agg import (
    FigureCanvasGTK3Agg as FigureCanvas)
from matplotlib.cm import get_cmap
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from ab12phylo import repo
from ab12phylo_cmd.filter import chmod_x
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 4


class gbl_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.view_gbl.set_model(data.gbl_model)
        iface.view_gbl.append_column(Gtk.TreeViewColumn(
            title='id', cell_renderer=Gtk.CellRendererText(), text=0))
        iface.view_gbl.connect('check-resize', self.get_height_resize, iface.gbl_spacer,
                               [iface.gbl_left, iface.gbl_right])

        iface.gbl_preset.set_id_column(0)
        iface.gaps.set_id_column(0)

        iface.tempspace.bak_ignore = set()
        iface.tempspace.params = list()
        for w_name in ['conserved', 'flank', 'good_block', 'bad_block']:
            wi = iface.__getattribute__(w_name)
            iface.tempspace.__setattr__(w_name, wi.get_adjustment())
        for wi in [iface.conserved, iface.flank]:
            wi.connect_after('changed', self.edit)  # when editing field, check adjustments

        iface.gaps.connect('changed', lambda *args: data.gbl.  # no assignment in lambda -> use __setattr__
                           __setattr__('gaps', iface.gaps.get_active_text()))
        iface.gbl_preset.connect_after('changed', self.re_preset)
        iface.gbl_preset.set_active_id('balanced')

        sel = iface.view_gbl.get_selection()
        sel.set_mode(Gtk.SelectionMode.MULTIPLE)
        sel.connect('changed', self.keep_visible, iface.view_gbl,
                    iface.parallel_gbl.props.vadjustment.props, iface.tempspace)
        iface.view_gbl.connect('key_press_event', self.delete_and_ignore_rows,
                               PAGE, sel)  # in-preview deletion
        iface.gbl_eventbox.connect('button_press_event', self.select_seqs, PAGE, iface.zoomer,
                                   iface.view_gbl, iface.tempspace)  # in-preview selection
        iface.gbl_eventbox.connect('scroll-event', self.xy_scale, PAGE)  # zooming

    def refresh(self):
        """Cause the initial plotting or load .png images for good responsiveness later on"""
        data = self.data
        iface = self.iface

        if not data.gene_ids:
            return

        # start the first re-plotting
        if 0 in data.msa_shape or not (self.wd / repo.PATHS.left).exists() \
                or not (self.wd / repo.PATHS.right).exists():
            # fetch the MSA shape
            data.gbl_model.clear()
            shared_ids = sorted(set.intersection(*data.gene_ids.values()))
            [data.gbl_model.append([_id]) for _id in shared_ids]
            i = len(shared_ids)
            data.msa_shape = [sum(data.msa_lens), i, -1, i]  # width-height-width_after-height
            iface.msa_shape.set_text('%d : %d' % tuple(data.msa_shape[:2]))
            iface.msa_shape_trimmed.set_text('-1 : %d' % i)
            # set the adjustment boundaries of the spin buttons
            self.re_preset(iface.gbl_preset)
            self.start_gbl()
            return

        # else load the pre-existing PNGs
        try:
            iface.msa_shape.set_text('%d : %d' % tuple(data.msa_shape[:2]))
            iface.msa_shape_trimmed.set_text(
                '%d : %d' % (data.msa_shape[2] - (len(data.genes) - 1) * len(repo.SEP),
                             data.msa_shape[3]))
        except TypeError:
            pass
        # place the existing png
        x_ratio = data.msa_shape[2] / data.msa_shape[0]
        try:
            self.load_image(iface.zoomer, PAGE, iface.gbl_left_vp, self.wd / repo.PATHS.left,
                            w=data.gbl_shape[0] * self.get_hadj(),
                            h=data.gbl_shape[1])
            self.load_image(iface.zoomer, PAGE, iface.gbl_right_vp, self.wd / repo.PATHS.right,
                            w=data.gbl_shape[0] * self.get_hadj() * x_ratio,
                            h=data.gbl_shape[1])
            self.load_colorbar(iface.palplot2)
            self.re_preset(iface.gbl_preset)
            self.win.show_all()
        except ValueError:
            data.msa_shape[3] = 0
            self.refresh()

    def reload_ui_state(self):
        data = self.data
        iface = self.iface
        iface.gbl_preset.set_active_id(data.gbl.preset)
        iface.gaps.set_active_id(data.gbl.gaps)
        for w_name in ['conserved', 'flank', 'good_block', 'bad_block']:
            p = iface.tempspace.__getattribute__(w_name).props
            p.value, p.lower, p.upper = data.gbl.__getattribute__(w_name)
            # configure(value, lower, upper, step-increment=1, page-increment=0, page-size=0)

    def re_preset(self, gbl_preset):
        """
        Apply a preset on toggling it.
        :param gbl_preset: the GtkComboBoxText with the presets
        """
        mode = gbl_preset.get_active_text()
        g = self.iface.tempspace
        n_sites, n_seqs = self.data.msa_shape[:2]

        assert type(n_sites) == int
        # block / un-block SpinButtons
        [self.iface.__getattribute__(w_name).set_sensitive(mode != 'skip')
         for w_name in ['conserved', 'flank', 'good_block', 'bad_block', 'gaps']]

        if mode == 'skip':
            return
        elif mode == 'relaxed':
            conserved = n_seqs // 2 + 1
            flank = conserved
            gaps = 'half'
            good_block = 5
            bad_block = 8
        elif mode == 'balanced':
            conserved = n_seqs // 2 + 1
            flank = min(n_seqs // 4 * 3 + 1, n_seqs)
            gaps = 'half'
            good_block = 5
            bad_block = 8
        elif mode == 'default':
            conserved = n_seqs // 2 + 1
            flank = min(int(n_seqs * 0.85) + 1, n_seqs)
            gaps = 'none'
            good_block = 10
            bad_block = 8
        elif mode == 'strict':
            conserved = int(n_seqs * .9)
            flank = conserved
            gaps = 'none'
            good_block = 10
            bad_block = 8
        else:
            raise RuntimeWarning(f'illegal mode: {mode}')

        # configure(value, lower, upper, step-increment=1, page-increment=0, page-size=0)
        g.conserved.configure(conserved, n_seqs // 2 + 1, n_seqs, 1, 0, 0)
        g.flank.configure(flank, n_seqs // 2 + 1, n_seqs, 1, 0, 0)
        g.good_block.configure(good_block, 2, n_sites, 1, 0, 0)
        g.bad_block.configure(bad_block, 0, n_sites, 1, 0, 0)
        self.iface.gaps.set_active_id(gaps)

    def edit(self, wi):
        """Just make sure flank is never smaller than conserved"""
        LOG.debug('editing')
        adj = wi.get_adjustment()
        flank = self.iface.tempspace.flank
        conserved = self.iface.tempspace.conserved
        if adj == flank:
            return  # break some loops
        # conserved has changed
        flank.configure(max(adj.get_value(), flank.get_value()),  # value
                        adj.get_value(), flank.get_upper(), 1, 0, 0)  # the others

    def start_gbl(self, run_after=None, force=False):
        """Set-up the Gblocks thread. This cannot be reached if Gblocks shall be skipped"""
        data = self.data
        iface = self.iface
        if not data.genes:
            LOG.debug('abort Gblocks')
            return
        elif iface.thread.is_alive():
            self.show_notification('Busy', secs=1)
            return
        elif run_after and (self.wd / repo.PATHS.msa).exists() \
                and not self.get_errors(PAGE):
            self.set_changed(PAGE, False)
            [do_func() for do_func in run_after]
            return

        iface.refresh.grab_focus()  # trick to make the adjustments realize they've changed
        if iface.gbl_preset.get_active_text() != 'skip':
            self.set_changed(PAGE, True)
            # get arguments from adjustments values
            g = iface.tempspace  # the param Namespace
            # do NOT swap order, do NOT swap order! Gblocks wants them in the wrong order!
            flank, cons, good, bad = [adj.get_value() for adj in
                                      [g.flank, g.conserved, g.good_block, g.bad_block]]
            gaps = iface.gaps.get_active_text()[0]
            # failsafe
            if flank < cons:
                flank = cons
            params = [flank, cons, bad, good, gaps]
            # check if they have changed
            if iface.tempspace.params == params and not force \
                    and iface.tempspace.bak_ignore == data.gbl.ignore_ids \
                    and not self.get_changed(PAGE):
                self.show_notification('MSA already trimmed with these parameters', secs=2)
                return
            iface.tempspace.params = params
        data.gbl_model.clear()

        # save_ui_state from iface.tempspace into data.gbl
        data.gbl.preset = iface.gbl_preset.get_active_text()
        data.gbl.gaps = iface.gaps.get_active_text()
        for w_name in ['conserved', 'flank', 'good_block', 'bad_block']:
            p = iface.tempspace.__getattribute__(w_name).props
            data.gbl.__setattr__(w_name, [p.value, p.lower, p.upper])

        iface.thread = threading.Thread(target=self._do_gbl)
        iface.run_after = run_after
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        sleep(.1)
        # return to main loop

    def _do_gbl(self):
        """Run the Gblocks thread"""
        data = self.data
        iface = self.iface

        iface.frac = .05
        iface.i = 0
        iface.k = len(data.genes) + 6

        if iface.gbl_preset.get_active_text() != 'skip':
            iface.text = 'prep'
            LOG.debug(iface.text)
            # look for local system Gblocks
            binary = shutil.which('Gblocks')
            local = bool(binary)
            # else pick deployed Gblocks
            binary = Path(binary) if binary else next(repo.TOOLS.rglob(f'Gblocks{os.getenv("PATHEXT", default="")}'))
            # Make the file executable
            chmod_x(binary)
            LOG.info('%s Gblocks' % ('local' if local else 'packaged'))

            # create base call -t=d sets the mode to nucleotides ... adapt?
            # leave -b2 and -b1 in this wrong order!
            arg = '%s %s -t=d -b2=%d -b1=%d -b3=%d -b4=%d -b5=%s -e=".txt" -s=y -p=n -k=y; exit 0' \
                  % tuple([binary, '"%s"'] + iface.tempspace.params)
            if sys.platform in ['win32', 'cygwin']:
                arg += ' & exit /b 0'
            else:
                arg += '; exit 0'
            LOG.debug(arg)
        else:
            arg = ''

        # fetch IDs
        shared_ids = sorted(set.intersection(*data.gene_ids.values()) - data.gbl.ignore_ids)
        [data.gbl_model.append([_id]) for _id in shared_ids]

        iface.i += 1
        GObject.idle_add(self._do_gbl2, shared_ids, arg)
        sleep(.1)
        return True

    def _do_gbl2(self, shared_ids, arg):
        """Bounce to idle to size the id column"""
        data = self.data
        iface = self.iface
        iface.thread.join()

        gbar, barspace = (True, 50) if len(data.genes) > 1 else (False, 10)

        iface.view_gbl.set_margin_bottom(barspace)
        iface.view_gbl.realize()
        sleep(.1)
        data.gbl_shape[1] = self.get_height_resize(iface.view_gbl, None, iface.gbl_spacer,
                                                   [iface.gbl_left, iface.gbl_right]) + barspace

        # re-direct to thread
        iface.thread = threading.Thread(target=self._do_gbl3, args=[shared_ids, arg, gbar, barspace])
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()

    def _do_gbl3(self, shared_ids, arg, gbar, barspace):
        data = self.data
        iface = self.iface

        errors = list()
        msa_lens = list()
        blocks = list()
        array = np.empty(shape=(len(shared_ids), 0), dtype=int)

        for gene in data.genes:
            iface.text = '%s: read MSA' % gene
            LOG.debug(iface.text)
            raw_msa = (self.wd / gene / ('%s_raw_msa.fasta' % gene)).resolve()
            msa = self.wd / gene / ('%s_msa.fasta' % gene)
            records = {r.id: r for r in SeqIO.parse(raw_msa, 'fasta')}
            take_out = {_id for _id in records.keys() if _id in data.gbl.ignore_ids}
            take_out = {_id: records.pop(_id) for _id in take_out}
            if take_out:
                # write newly dropped sequences to backup file
                new_take_out = {_id: r for _id, r in take_out.items()
                                if _id not in iface.tempspace.bak_ignore}
                with open(self.wd / gene / ('%s_raw_msa_dropped.fasta' % gene), 'a') as fasta:
                    SeqIO.write(new_take_out.values(), fasta, 'fasta')
                # overwrite MSA without all dropped samples
                with open(raw_msa, 'w') as fasta:
                    SeqIO.write(records.values(), fasta, 'fasta')
            if sorted(records.keys()) != shared_ids:
                errors.append('MSA for %s does not match the dataset, please re-build.' % gene)
                sleep(.1)
                GObject.idle_add(self.stop_gbl, errors)
                return True
            ar = np.array([repo.seqtoint(records[_id].seq.upper()) for _id in shared_ids])
            msa_lens.append(ar.shape[1])
            # get the array columns that are not only gaps
            usable_sites = [i for i in range(ar.shape[1]) if set(ar[:, i]) != {repo.toint('-')}]
            array = np.hstack((array, ar,))  # shared.SEP would need to be stacked here
            del ar
            data.msa_shape[:2] = array.shape[::-1]

            if iface.gbl_preset.get_active_text() == 'skip':
                LOG.debug('skipping %s' % gene)
                shutil.copy(raw_msa, msa)
                iface.i += 1
                continue

            iface.text = '%s: run Gblocks' % gene
            LOG.debug(iface.text)
            with open(self.wd / gene / 'gblocks.log', 'w') as log_handle:
                try:
                    LOG.debug(arg % raw_msa)
                    subprocess.run(arg % raw_msa, shell=True, check=True,
                                   stdout=log_handle, stderr=log_handle)
                except (OSError, subprocess.CalledProcessError) as e:
                    errors.append(str(e))
                    log_handle.write(str(e))
                    continue

            # parse result
            iface.text = '%s: parse result' % gene
            LOG.debug(iface.text)

            shutil.move(raw_msa.with_suffix('.fasta.txt'), msa)
            # get the good blocks from the last pseudo-sequence in the text mask file
            for pseudo_seq in SeqIO.parse(raw_msa.with_suffix('.fasta.txtMask'), 'fasta'):
                mask = pseudo_seq
            # map the Gblocks mask to the original MSA sites
            line_blocks = [usable_sites[i] for i, char in enumerate(mask.seq) if char == '#']
            LOG.debug(line_blocks)
            if not line_blocks:
                err = '%s: no good blocks' % gene
                LOG.error(err)
                errors.append(err)
                continue

            shift = sum(msa_lens[:-1])
            blocks.extend([i + shift for i in line_blocks])
            iface.i += 1
        data.msa_lens = msa_lens

        iface.text = 'concatenating MSAs'
        iface.tempspace.bak_ignore = {i for i in data.gbl.ignore_ids}
        # if 'aligner' not in iface:
        iface.aligner, cmd = self.get_msa_build_cmd(
            repo.toalgo(iface.msa_algo.get_active_text()), self.wd, data.genes)
        # iface.aligner.reset_paths(self.wd, self.wd / repo.PATHS.msa)
        data.msa_shape[2], data.msa_shape[3] = iface.aligner.concat_msa(gui=shared_ids)
        iface.text = 'computing SHA256 hash'
        LOG.debug(iface.text)
        self.get_hashes(repo.PATHS.msa, PAGE)

        iface.i += 1
        iface.text = 'plot MSAs'
        LOG.debug(iface.text)

        data.gbl_shape[0] = array.shape[1] * self.get_hadj()
        # make gaps transparent
        array = np.ma.masked_where(array > repo.toint('else'), array)

        # create a transparency mask
        gbl_mask = np.full(array.shape, repo.ALPHA)
        gbl_mask[:, blocks] = 1

        data.msa_shape[2] = len(blocks)  # for completeness
        LOG.debug('msa shape: %s' % str(data.msa_shape))
        x_ratio = data.msa_shape[2] / data.msa_shape[0]
        LOG.debug('x ratio: %.3f' % x_ratio)

        # adjust maximum size
        scale = 6
        while max(data.msa_shape[:2]) * scale > 2 ** 14:
            scale -= 1
        LOG.debug('scaling gbl with %d' % scale)

        if iface.gbl_preset.get_active_text() != 'skip':
            for alpha, blocks, gtk_bin, png_path, x_ratio, width \
                    in zip([gbl_mask, 1], [range(array.shape[1]), blocks],
                           [iface.gbl_left_vp, iface.gbl_right_vp],
                           [repo.PATHS.left, repo.PATHS.right], [1, x_ratio],
                           [data.msa_shape[0], data.msa_shape[2]]):
                f = Figure()  # figsize=(width / shared.DPI, data.msa_shape[1] / shared.DPI), dpi=shared.DPI)  # figaspect(data.msa_shape[1] / width))
                f.set_facecolor('none')
                f.set_figheight(data.msa_shape[1] / repo.DPI * 5)
                f.set_figwidth(max(1, width) / repo.DPI)
                # f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)

                # leave room at the bottom for the gene legend bar
                b = barspace / data.gbl_shape[1]
                ax = f.add_axes([0, b, 1, 1 - b])

                mat = ax.matshow(array[:, blocks], alpha=alpha, cmap=ListedColormap(repo.colors),
                                 vmin=-.5, vmax=len(repo.colors) - .5, aspect='auto')
                if not gbar:
                    LOG.debug('adding ticks')
                    [ax.spines[t].set_visible(False) for t in ['left', 'right', 'top', 'bottom']]
                    ax.yaxis.set_visible(False)
                    ax.xaxis.set_ticks_position('bottom')
                    ax.tick_params(colors=iface.FG, pad=.2, length=0, labelsize=1)
                    ax.xaxis.set_ticks([i for i in range(1, array[:, blocks].shape[1] - 1) if not i % 100])
                else:
                    ax.axis('off')
                    LOG.debug('adding gene marker bar')
                    # build matrix of gene indicators
                    gm = [[i] * l for i, l in enumerate(msa_lens)]
                    if x_ratio == 1:
                        # add the spacer
                        for i in range(len(data.genes) - 1):
                            gm[i] += [-1] * len(repo.SEP)
                    gm = [gi for gl in gm for gi in gl]
                    # turn into np array
                    gm = np.vstack([np.array(gm)] * 2)
                    # make spacer transparent
                    gm = np.ma.masked_where(gm < 0, gm)
                    # trim the array early
                    gm = gm[:, blocks]
                    gene_colors = get_cmap('GnBu', len(data.genes))

                    # plot the marker bar onto the MSA graphic
                    bax = f.add_axes([0, b * .4, 1, b / 3])
                    bar = bax.pcolormesh(gm, cmap=gene_colors)
                    # bax.axis('off')
                    [bax.spines[t].set_visible(False) for t in ['left', 'right', 'top', 'bottom']]
                    bax.yaxis.set_visible(False)
                    bax.xaxis.set_ticks_position('bottom')
                    bax.tick_params(colors=iface.FG, pad=.2, length=0, labelsize=1)
                    bax.xaxis.set_ticks([i for i in range(1, gm.shape[1] - 1) if not i % 100])

                    # plot a legend for the gene marker bar
                    with plt.rc_context({'axes.edgecolor': iface.FG, 'xtick.color': iface.FG}):
                        iface.text = 'gene marker bar'
                        LOG.debug(iface.text)
                        Path.mkdir(self.wd / repo.PATHS.phylo_msa.parent, exist_ok=True)
                        fig = plt.figure(figsize=(4, .2))
                        cax = fig.add_subplot(111)
                        cbar = ColorbarBase(
                            ax=cax, cmap=gene_colors, orientation='horizontal',
                            ticks=[(.5 / len(data.genes) + j * 1 / len(data.genes))
                                   for j in range(len(data.genes))])
                        cbar.ax.set_xticklabels(data.genes)
                        fig.savefig(self.wd / repo.PATHS.gbar, transparent=True,
                                    bbox_inches='tight', pad_inches=0, dpi=600)
                        plt.close(fig)
                        del fig, cbar

                iface.i += 1
                iface.text = 'save PNG'
                LOG.debug(iface.text)
                Path.mkdir(self.wd / png_path.parent, exist_ok=True)
                f.savefig(self.wd / png_path, transparent=True,
                          dpi=scale * repo.DPI, bbox_inches='tight', pad_inches=0.00001)

                if iface.rasterize.props.active:
                    iface.text = 'place PNG'
                    LOG.debug(iface.text)
                    self.load_image(iface.zoomer, PAGE, gtk_bin, self.wd / png_path,
                                    data.gbl_shape[0] * x_ratio * self.get_hadj(), data.gbl_shape[1])
                else:
                    iface.text = 'place vector'
                    LOG.debug(iface.text)
                    canvas = FigureCanvas(f)
                    canvas.set_size_request(max(len(blocks) * self.get_hadj(), -1),
                                            data.gbl_shape[1])  # width, height
                    try:
                        ch = gtk_bin.get_child()
                        if ch:
                            gtk_bin.remove(ch)
                        gtk_bin.add(canvas)
                    except Exception as ex:
                        LOG.error(ex)
                iface.i += 1

                gtk_bin.realize()
                gtk_bin.show_all()

        # re-size
        LOG.debug('re-sizing again')
        for wi in [iface.gbl_left, iface.gbl_right]:
            wi.set_max_content_height(data.gbl_shape[1])

        self.load_colorbar(iface.palplot2)

        if gbar:
            self.load_colorbar(iface.gbar1, gbar=True)
            iface.gbar1.set_visible(True)
        else:
            iface.gbar1.set_visible(False)

        iface.text = 'idle'
        iface.frac = 1
        sleep(.1)
        GObject.idle_add(self.stop_gbl, errors)
        return True

    def stop_gbl(self, errors):
        """Finish the Gblocks thread"""
        data, iface = self.data, self.iface
        iface.thread.join()
        LOG.info('gbl thread idle')
        sleep(.1)
        self.update(PAGE)
        self.win.show_all()
        self.set_errors(PAGE, bool(errors))
        self.set_changed(PAGE, False)
        iface.msa_shape.set_text('%d : %d' % tuple(data.msa_shape[:2]))
        iface.msa_shape_trimmed.set_text(
            '%d : %d' % (data.msa_shape[2], data.msa_shape[3]))

        # possibly re-configure flank and conserved
        g = iface.tempspace
        n_seqs = data.msa_shape[1]
        min_c = n_seqs // 2 + 1
        if g.flank.props.upper != n_seqs:
            g.conserved.configure(max(min_c, g.conserved.props.value), min_c, n_seqs, 1, 0, 0)
            g.flank.configure(max(min_c, g.flank.props.value), min_c, n_seqs, 1, 0, 0)
        if errors:
            self.show_notification('Errors during MSA trimming', errors)
            return
        if iface.run_after:
            [do_func() for do_func in iface.run_after]
        return

    def undrop_seqs(self):
        """Write dropped sequences back to the untrimmed MSAs."""
        ignore_ids, data, iface = self.data.gbl.ignore_ids, self.data, self.iface
        for gene in data.genes:
            dropped_file = self.wd / gene / ('%s_raw_msa_dropped.fasta' % gene)
            if not dropped_file.is_file():
                continue
            put_in = {r.id: r for r in SeqIO.parse(dropped_file, 'fasta')}

            unknown = {_id for _id in put_in.keys() if _id not in ignore_ids}
            assert not unknown, 'Found wrong dropped sequences: %s' % ','.join(unknown)

            with open(self.wd / gene / ('%s_raw_msa.fasta' % gene), 'a') as fasta:
                SeqIO.write(put_in.values(), fasta, 'fasta')
            dropped_file.unlink()

        # save in metadata
        for _id in ignore_ids:
            for gene, genemeta in data.metadata.items():
                if _id in genemeta:
                    genemeta[_id]['quality'] = ''
        self.write_metadata()

        ignore_ids.clear()
        self.iface.tempspace.bak_ignore.clear()
        self.start_gbl(force=True)

    def drop_seqs(self):
        """
        On manually deleting samples from the trimmed MSA, overwrite 'old'
        trimmed MSA file, save in and write metadata.
        """
        ignore_ids, data, iface = self.data.gbl.ignore_ids, self.data, self.iface
        if not ignore_ids:
            return
        # overwrite old MSA
        with open(self.wd / (repo.PATHS.msa + '_TEMP'), 'w') as fasta:
            for record in SeqIO.parse(self.wd / repo.PATHS.msa, 'fasta'):
                if record.id not in ignore_ids:
                    SeqIO.write(record, fasta, 'fasta')
        LOG.debug('dropped %d sequences from trimmed MSA' % len(ignore_ids))
        shutil.move(self.wd / (repo.PATHS.msa + '_TEMP'), self.wd / repo.PATHS.msa)

        # save in metadata
        for _id in ignore_ids:
            for gene, genemeta in data.metadata.items():
                if _id in genemeta:
                    genemeta[_id]['quality'] = 'manually dropped at trim_MSA'
        self.write_metadata()



# 2021 Leo Kaindl

import logging
import shutil
import subprocess
import sys
import threading
from pathlib import Path

import gi
from Bio import SeqIO

gi.require_version('Gtk', '3.0')
from gi.repository import GObject

from ab12phylo import repo
from ab12phylo.gtk_base import ab12phylo_app_base

LOG = logging.getLogger(__name__)
PAGE = 3


class msa_page(ab12phylo_app_base):

    def __init__(self):
        super().__init__()
        data = self.data
        iface = self.iface

        iface.msa_algo.set_entry_text_column(0)
        iface.msa_algo.set_id_column(0)
        iface.remote_algo.set_id_column(0)

        iface.msa_cmd.connect('focus_out_event', lambda widget, *args: data.msa.cmd.update(
            {repo.toalgo(iface.msa_algo.get_active_text()): widget.get_buffer().props.text.strip()}))
        iface.remote_cmd.connect('focus_out_event', lambda widget, *args: data.msa.remote_cmd.update(
            {repo.toalgo(iface.remote_algo.get_active_text()): widget.get_buffer().props.text.strip()}))

        iface.msa_algo.connect('changed', self.get_help)
        iface.remote_algo.connect('changed', self.get_help, True)
        iface.msa_import.connect('file-set', self.load_msa)
        iface.msa_exe.connect('file-set', self.get_help, True, True)

        # connect buttons
        iface.msa_build.connect('clicked', self.start_align)
        self.bind_accelerator(self.accelerators, iface.msa_build, 'Return')
        iface.remote_build.connect('clicked', self.start_align, True)
        self.bind_accelerator(self.accelerators, iface.remote_build, 'Return')

        data.msa.stack_child_name = iface.align_stack.get_visible_child_name()

    def get_help(self, widget, remote=False, try_path=False):
        data = self.data
        iface = self.iface

        if not data.genes:
            return
        if remote:
            data.msa.algo = repo.toalgo(iface.remote_algo.get_active_text())
            client = repo.TOOLS / 'MSA_clients' / (data.msa.algo + '.py')
            self.set_helpers('%s %s ' % (sys.executable, client), iface.remote_help,
                             data.msa.remote_cmd, data.msa.algo, True, iface.remote_cmd)
        else:
            data.msa.algo = repo.toalgo(iface.msa_algo.get_active_text())
            exe = shutil.which(data.msa.algo)
            exe = widget.get_active_text() if try_path else exe
            if exe:
                # get the --help output and save it in the lookup field on the right
                iface.msa_exe.set_filename(exe)
                self.set_helpers('%s --help; exit 0' % exe, iface.msa_help,
                                 data.msa.cmd, data.msa.algo, False, iface.msa_cmd)
            else:
                # no executable found; unselect in path box
                iface.msa_exe.unselect_all()
                txt = data.msa.algo + ' was not found on your system $PATH. You can try ' \
                                      'manually specifying the path to the executable.'
                iface.msa_help.get_buffer().props.text = txt  # show this snarky line above
                iface.msa_cmd.get_buffer().props.text = ''  # no cmd suggestion

    def set_helpers(self, cmdline, help_view, help_dict, algo, remote, cmd_view):
        proc = subprocess.Popen(cmdline, shell=True,
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = proc.communicate()
        txt = out.decode().strip() + '\n\n' + err.decode().strip()
        help_view.get_buffer().props.text = txt.strip()

        # get the suggested command and allow user modification in the left field
        txt = help_dict.get(algo, '')  # fetch saved
        if txt == '' or 'aligner' not in self.iface:  # deleting all content will also get you back the original
            self.iface.aligner, txt = self.get_msa_build_cmd(algo, self.wd, self.data.genes, remote)  # fetch new
            help_dict[algo] = txt  # save

        cmd_view.get_buffer().props.text = help_dict[algo]  # show

    def refresh_paths(self):
        LOG.debug('refresh_paths')
        old = str(self.iface.aligner.dir)
        self.iface.aligner.reset_paths(self.wd, self.wd / repo.PATHS.msa)
        for algo, txt in self.data.msa.cmd.items():
            self.data.msa.cmd[algo] = txt.replace(old, str(self.wd))
        # also replace it in the GtkTextView
        bf = self.iface.msa_cmd.get_buffer()
        bf.props.text = bf.props.text.replace(old, str(self.wd))

    def start_align(self, widget, remote=False, run_after=None):
        """
        Starts an MSA building thread unless one of the following conditions is met:
        a) another thread is active -> abort + forbid proceeding.
        b) this function was called from the _Next button and an MSA already exists
        -> accept + allow proceeding.
        c) there were no changes registered for this page and no proceeding (not
        called by _Next) -> accept + show notification.

        :param widget: required for callback, ignored
        :param remote: if the MSA should be constructed using the EMBL-EBI API
        :param run_after: the function to run afterwards; usually flip to next page
        :return:
        """
        data = self.data
        iface = self.iface
        if iface.thread.is_alive():  # a)
            self.show_notification('Busy', secs=1)
            return
        elif data.msa.last_cmd == [data.msa.cmd, data.msa.remote_cmd][remote][data.msa.algo] \
                and not self.get_changed(PAGE):
            self.show_notification('MSA already generated, please proceed')
            return
        elif all([(self.wd / gene / ('%s_raw_msa.fasta' % gene)).exists() for gene in data.genes]) \
                and run_after and not self.get_errors(PAGE):  # b)
            self.set_changed(PAGE, False)
            [do_func() for do_func in run_after]
            return
        if 'aligner' not in iface:
            self.get_help(None, remote)
        self.save_ui_state()
        self.save()

        data.msa_lens.clear()
        iface.align_stack.props.sensitive = False
        iface.thread = threading.Thread(target=self.do_align, args=[remote])
        iface.run_after = run_after
        GObject.timeout_add(100, self.update, PAGE)
        iface.thread.start()
        return
        # return to main loop

    def do_align(self, remote=False):
        data = self.data
        iface = self.iface
        errors = list()
        iface.frac = .05
        iface.i = 0
        iface.k = len(data.genes)
        funcs, arg_dicts = [iface.aligner.build_local, iface.aligner.build_remote], \
                           [data.msa.cmd, data.msa.remote_cmd]
        try:
            for gene in data.genes:
                iface.text = 'aligning %s [%d/%d]' % (gene, iface.i + 1, iface.k)
                LOG.debug(iface.text)
                arg = arg_dicts[remote][data.msa.algo]
                data.msa.last_cmd = arg
                try:
                    funcs[remote](gene, new_arg=
                    arg % tuple([gene] * (4 - remote)))  # interpreting bool as int here
                except (FileNotFoundError, subprocess.CalledProcessError):
                    self.refresh_paths()
                    # try again once more
                    funcs[remote](gene, new_arg=arg % tuple([gene] * 4))
                # fetch MSA length
                for r in SeqIO.parse(self.wd / gene / ('%s_raw_msa.fasta' % gene), 'fasta'):
                    data.msa_lens.append(len(r))
                    break
                iface.i += 1
            iface.frac = 1
            iface.text = 'idle'
        except (OSError, subprocess.CalledProcessError) as e:
            errors.append('%s at task %d (%s). invalid command?'
                          % (type(e), iface.i, iface.text))
        except FileNotFoundError:
            errors.append('MSA/sequences file not found. Did you just save somewhere new?')
        except TypeError:
            raise ValueError('replace string formatting with something smarter')
        GObject.idle_add(self.stop_align, errors)

    def stop_align(self, errors):
        iface = self.iface
        iface.thread.join()
        iface.align_stack.props.sensitive = True
        self.win.show_all()
        LOG.info('msa thread idle')
        self.set_errors(PAGE, bool(errors))
        self.set_changed(PAGE, False)
        if errors:
            self.show_notification('Errors during MSA building', errors)
            return
        if iface.run_after:
            [do_func() for do_func in iface.run_after]
        else:
            self.show_notification('MSA building finished', items=None)
        return

    def load_msa(self, widget):
        data = self.data
        iface = self.iface
        try:
            Path.mkdir(self.wd / repo.PATHS.import_msa.parent, exist_ok=True)
            shutil.copy(widget.get_filename(), self.wd / repo.PATHS.import_msa)
        except shutil.SameFileError:
            pass
        except Exception as ex:
            self.show_notification(str(ex))
            LOG.error(ex)
        self.get_hashes(repo.PATHS.import_msa, PAGE)
        data.genes = ['import']
        data.gene_ids = {'import': {r.id for r in SeqIO.parse(
            self.wd / repo.PATHS.import_msa, 'fasta')}}
        iface.aligner, cmd = self.get_msa_build_cmd(
            repo.toalgo(iface.msa_algo.get_active_text()), self.wd, data.genes)
        # write a metadata.tsv
        with open(self.wd / repo.PATHS.tsv, 'w') as metadata:
            metadata.write('id\tgene\n')
            for _id in data.gene_ids['import']:
                metadata.write('%s\timport\n' % _id)
        LOG.debug('using imported MSA')
        self.set_changed(PAGE, False)
        self.save_ui_state()

    def refresh(self):
        if self.iface.align_stack.get_visible_child_name() == 'local':
            self.get_help(None)
        elif self.iface.align_stack.get_visible_child_name() == 'remote':
            self.get_help(None, remote=True)

    def save_ui_state(self):
        ns = self.data.msa
        iface = self.iface
        ns.stack_child_name = iface.align_stack.get_visible_child_name()
        ns.msa_algo_id = iface.msa_algo.get_active_id()
        ns.msa_exe_filename = iface.msa_exe.get_filename()
        ns.remote_algo_id = iface.remote_algo.get_active_id()
        ns.msa_import_filename = iface.msa_import.get_filename()

    def reload_ui_state(self):
        ns = self.data.msa
        iface = self.iface
        iface.align_stack.set_visible_child_name(ns.stack_child_name)
        iface.msa_algo.set_active_id(ns.msa_algo_id)
        if ns.msa_exe_filename:
            iface.msa_exe.set_filename(ns.msa_exe_filename)
        iface.remote_algo.set_active_id(ns.remote_algo_id)
        if ns.msa_import_filename:
            iface.msa_import.set_filename(ns.msa_import_filename)

        iface.msa_cmd.get_buffer().props.text = \
            ns.cmd.get(repo.toalgo(ns.msa_algo_id), '')
        iface.remote_cmd.get_buffer().props.text = \
            ns.remote_cmd.get(repo.toalgo(ns.remote_algo_id), '')
